#include <cstdint>
#include <cstdio>
#include <memory>
#include <cstring>

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;

#define UINT24_MAX ((1 << 24) - 1)

#define CODE_FLASH_ERASE_SIZE 0x400
#define CODE_FLASH_PAGES_TO_BYTES(x) ((x) * CODE_FLASH_ERASE_SIZE)

struct EntryHeader {
    enum Magic : u8 {
        Word0 = 0xa5,
        Word1 = 0xc3,
    };
    u32 words[2];
    bool has_word0_magic() const {
        return (words[0] & UINT8_MAX) == Magic::Word0;
    }
    u16 index() const {
        return (words[0] >> 8) & UINT16_MAX;
    }
    u16 link() const {
        return (words[0] >> 24) & UINT8_MAX;
    }
    u32 write_counter() const {
        return words[1] & UINT24_MAX;
    }
    bool has_word1_magic() const {
        return ((words[1] >> 24) & UINT8_MAX) == Magic::Word1;
    }
    bool has_magics() const {
        return has_word0_magic() && has_word1_magic();
    }
};
static_assert(sizeof(EntryHeader) == 8, "sizeof(EntryHeader)");

struct DataEntry {
    EntryHeader hdr;
    u8 data[8];
};
static_assert(sizeof(DataEntry) == 16, "sizeof(DataEntry)");

// Original code explicitly lists offsets + sizes of everything.
// This code assumes everything becomes contiguous.

struct VolumeHeaderDesc {
    u16 len;
    u32 count;
    u32 total_len() const {
        return len * count;
    }
};

struct DataRegionDesc {
    u16 flat_len;
    u16 sparse_len;
    u32 count;
    u32 single_len() const {
        return flat_len + sparse_len;
    }
    u32 total_len() const {
        return single_len() * count;
    }
};

struct NvsInfo {
    const char *name;
    u32 offset;
    VolumeHeaderDesc volume_header;
    DataRegionDesc data_region;
    u32 total_len() const {
        return volume_header.total_len() + data_region.total_len();
    }
};

struct NvsParser {
    NvsParser(u8 *storage, const NvsInfo &info) : info_(info) {
        base = storage + info_.offset;
    }
    u8 *GetActiveVolume() const {
        return base + info_.volume_header.len * active_volume;
    }
    u32 VolumeNumEntries() const {
        return info_.volume_header.len / sizeof(EntryHeader) - 1;
    }
    u8 *GetActiveDataRegion() const {
        return base + info_.volume_header.total_len() +
            info_.data_region.single_len() * active_data_region;
    }
    u32 DataNumEntries() const {
        return info_.data_region.sparse_len / sizeof(DataEntry);
    }
    EntryHeader *GetVolumeHeader() const {
        return (EntryHeader *)GetActiveVolume();
    }
    EntryHeader *GetVolumeEntry(u32 index) const {
        return &GetVolumeHeader()[1 + index];
    }
    u8 *GetDataEntryFlat(u32 index) const {
        return GetActiveDataRegion() + index * 8;
    }
    DataEntry *GetDataEntrySparse(u32 index) const {
        auto sparse_base = GetActiveDataRegion() + info_.data_region.flat_len;
        return &((DataEntry *)sparse_base)[index];
    }
    EntryHeader *GetDataEntryHeader(u32 index) const {
        return &GetDataEntrySparse(index)->hdr;
    }
    u8 *GetDataEntry(u32 index) const {
        return GetDataEntrySparse(index)->data;
    }
    bool ReadVolumeWriteCounter(u16 *counter) {
        auto hdr = GetVolumeHeader();
        *counter = hdr->index();
        return hdr->has_magics();
    }
    bool ResolveLastValidVolumeEntry(u32 *index) {
        for (u32 i = VolumeNumEntries(); i > 0; i--) {
            u32 entry_index = i - 1;
            if (GetVolumeEntry(entry_index)->has_magics()) {
                *index = entry_index;
                return true;
            }
        }
        return false;
    }
    bool ResolveActive() {
        u16 volume_counter = 0;
        u32 last_active_volume = 0;
        for (active_volume = 0; active_volume < info_.volume_header.count; active_volume++) {
            u16 counter;
            if (ReadVolumeWriteCounter(&counter)) {
                u32 last_entry;
                if (ResolveLastValidVolumeEntry(&last_entry) && counter >= volume_counter) {
                    last_active_volume = active_volume;
                    active_volume_entry = last_entry;
                }
            }
        }
        active_volume = last_active_volume;
        auto data_ptr = GetVolumeEntry(active_volume_entry);
        active_data_region = data_ptr->link();
        printf("active volume %d. data region %d (counter starts @ %x)\n",
            active_volume, active_data_region, data_ptr->write_counter());
        return true;
    }
    u8 *ResolveLastValidData(u32 index) {
        for (u32 i = DataNumEntries(); i > 0; i--) {
            u32 entry_index = i - 1;
            auto entry = GetDataEntryHeader(entry_index);
            if (entry->has_magics() && entry->index() == index) {
                return GetDataEntry(entry_index);
            }
        }
        // just return the flattened data
        return GetDataEntryFlat(index);
    }
    std::unique_ptr<u8[]> ReadAll() {
        u32 flat_len = info_.data_region.flat_len;
        auto data = std::make_unique<u8[]>(flat_len);
        for (u32 offset = 0; offset < flat_len; offset += 8) {
            memcpy(&data[offset], ResolveLastValidData(offset / 8), 8);
        }
        return data;
    }
    const NvsInfo info_;
    u8 *base;
    u32 active_volume;
    u32 active_volume_entry;
    u32 active_data_region;
};

const NvsInfo snvs_info = {
    "sc_snvs",
    0,
    { CODE_FLASH_PAGES_TO_BYTES(1), 2 },
    { CODE_FLASH_PAGES_TO_BYTES(1), CODE_FLASH_PAGES_TO_BYTES(5), 8 },
};

const NvsInfo nvs_info = {
    "sc_nvs",
    0xe000,
    { CODE_FLASH_PAGES_TO_BYTES(1), 2 },
    { CODE_FLASH_PAGES_TO_BYTES(1), CODE_FLASH_PAGES_TO_BYTES(2), 2 },
};

#define NVS_BASE 0x60000

std::unique_ptr<u8[]> read_nvs(const char *path) {
    FILE *f = fopen(path, "rb");
    if (!f) {
        return nullptr;
    }
    if (fseek(f, NVS_BASE, SEEK_SET)) {
        return nullptr;
    }
    u32 nvs_len = nvs_info.offset + nvs_info.total_len();
    auto buf = std::make_unique<u8[]>(nvs_len);
    if (!fread(buf.get(), nvs_len, 1, f)) {
        buf = nullptr;
    }
    fclose(f);
    return buf;
}

int main(int argc, char **argv) {
    if (argc < 2) {
        return 1;
    }
    auto nvs = read_nvs(argv[1]);
    if (!nvs) {
        return 1;
    }
    for (auto &info : {snvs_info, nvs_info}) {
        auto parser = NvsParser(nvs.get(), info);
        if (!parser.ResolveActive()) {
            puts("ResolveActive");
            continue;
        }
        auto last_data = parser.ReadAll();
        FILE *f = fopen(info.name, "wb");
        fwrite(last_data.get(), info.data_region.flat_len, 1, f);
        fclose(f);
    }
    puts("done");
    return 0;
}
